/*
 * Copyright 2010, 2011, 2013 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "CarSelectionState.h"
#include "SurfaceBank.h"
#include "AIDriver.h"
#include "StateManager.h"
#include "RaceSetup.h"
#include <algorithm>


const Color titleColor = {0xFF, 0xDD, 0x10};

CarSelectionState::CarSelectionState(int humanPlayers, int computerPlayers, int laps, int tracks, bool tournament, bool carCollisions, int difficulty)
:	humanPlayers(humanPlayers),
	computerPlayers(computerPlayers),
	laps(laps),
	tracks(tracks),
	tournament(tournament),
	carCollisions(carCollisions),
	difficulty(difficulty),
	titleRO(Surface("Player 1, select your car!", BIG_FONT, titleColor), 80, 20),
	selectedCar(0)
{
	addRenderObject(titleRO);
	
	for (int i = 0; i < 8; i++)
	{
		cars[i].id = i;
		cars[i].taken = false;
		cars[i].renderObject.setSurface(SurfaceBank::getCar(i));
		cars[i].renderObject.moveTo(80 + 20 * i, 75);
		addRenderObject(cars[i].renderObject);
	}
}

void CarSelectionState::onKeyDown(SDLKey key)
{
	switch (key)
	{
	case SDLK_LEFT:
		prevCar();
		break;
	case SDLK_RIGHT:
		nextCar();
		break;
	case SDLK_RETURN:
	case SDLK_KP_ENTER:
		selectCar();
		break;
	case SDLK_ESCAPE:
		StateManager::getInstance().enterRacePreparationMenuState(false);
		break;
	default:
		break;
	}
}

void CarSelectionState::nextCar()
{
	cars[selectedCar].renderObject.setRotation(0.0);
	do
	{
		selectedCar = (selectedCar + 1) % 8;
	} while (cars[selectedCar].taken);
}
void CarSelectionState::prevCar()
{
	cars[selectedCar].renderObject.setRotation(0.0);
	do
	{
		selectedCar = (selectedCar > 0 ? (selectedCar - 1) % 8 : 7);
	} while (cars[selectedCar].taken);
}

void CarSelectionState::selectCar()
{
	playerCars.push_back(cars[selectedCar].id);
	cars[selectedCar].taken = true;
	if (playerCars.size() < (std::size_t) humanPlayers)
	{		
		titleRO.setSurface(Surface("Player " + int2string(playerCars.size() + 1) + ", select your car!", BIG_FONT, titleColor));
		cars[selectedCar].renderObject.setSurface(Surface());
		nextCar();
	}
}

void CarSelectionState::enterNextState()
{
	// give remaining cars to the computer players randomly
	for (int i = 0; i < 8; i++)
	{
		if (!cars[i].taken)
		{
			playerCars.push_back(cars[i].id);
		}
	}
	std::random_shuffle(playerCars.begin() + humanPlayers, playerCars.end());
	playerCars.resize(humanPlayers + computerPlayers);
	
	StateManager& sm = StateManager::getInstance();
	
	std::vector<Driver*> drivers;
	for (int i = 0; i < humanPlayers; i++)
	{
		drivers.push_back(sm.getDriver(i));
	}
	for (int i = 0; i < computerPlayers; i++)
	{
		drivers.push_back(&AIDriver::getInstance(difficulty));
	}
	
	StateManager::getInstance().enterRaceState(RaceSetup(drivers, playerCars, laps, tracks, tournament, carCollisions));
}


void CarSelectionState::update()
{
	static double rotation = 0.0;
	rotation += 0.07;
	cars[selectedCar].renderObject.setRotation(rotation);
	
	if (playerCars.size() == (std::size_t) humanPlayers)
	{
		enterNextState();
	}
}

