/*
 * Copyright 2013 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ControllerSettings.h"
#include "StateManager.h"
#include "Driver.h"
#include "InputDevice.h"
#include <cmath>

static std::string controlLabel[] = 
{
	"Accelerate: ",
	"Brake: ",
    "Left: ",
    "Right: "
};

const Color TITLE_COLOR = {0xFF, 0xDD, 0x10};

ControllerSettings::ControllerSettings()
:	player(1),
	device(0),
	returnToMainMenu(false),
	mode(-1)
{
	background.fill(120, 124, 224);
	Surface title("Controller settings", BIG_FONT, TITLE_COLOR);
	title.draw(background, 115, 30);
	
	menu.addItem("Player", 0, &player, 1, 8);
	StateManager& sm = StateManager::getInstance();
	std::vector<std::string> deviceNames = sm.getDeviceNames();
	menu.addItem("Device", 0, &device, deviceNames);
	
	std::string controlLabel[] = {"Accelerate: ",
	                              "Brake: ",
	                              "Left: ",
	                              "Right: "};
	for (int i = 0; i < 4; ++i)
	{
		enterControlMode[i] = false;
		menu.addItem(controlLabel[i], &enterControlMode[i]);
	}
	menu.addItem("Return to main menu", &returnToMainMenu);
	
	int maxDeviceName = 0;
	for (std::size_t i = 0; i < deviceNames.size(); ++i)
	{
		if (int(deviceNames[i].length()) > maxDeviceName)
		{
			maxDeviceName = deviceNames[i].length();
		}
	}
	int textLength = std::min(std::max(maxDeviceName + 8, 14), 50);
	int textWidth = textLength * 6; // most letters are 6 pixels
	menu.addRenderObjects(this, (GAME_WIDTH - textWidth) / 2, 50, 0, 10);
	
	
	for (int i = 0; i < 8; ++i)
	{
		drivers[i].resize(deviceNames.size());
		Driver* driver = sm.getDriver(i);
		int deviceIndex = driver->getDeviceIndex();
		drivers[i][deviceIndex] = driver;
		deviceIndices[i] = deviceIndex;
	}
	updatePlayer();
}

ControllerSettings::~ControllerSettings()
{
	StateManager& sm = StateManager::getInstance();
	for (int i = 0; i < 8; ++i)
	{
		for (std::size_t j = 0; j < drivers[i].size(); ++j)
		{
			if (sm.getDriver(i) != drivers[i][j])
			{
				delete drivers[i][j];
			}
		}
	}
}

void ControllerSettings::update()
{
	
}

void ControllerSettings::onKeyDown(SDLKey key)
{
	if (mode == -1)
	{
		int oldPlayer = player;
		int oldDevice = device;
		if (menu.onKeyDown(key))
		{
			StateManager& sm = StateManager::getInstance();
			if (returnToMainMenu)
			{
				for (int i = 0; i < 8; ++i)
				{
					for (std::size_t j = 0; j < drivers[i].size(); ++j)
					{
						if (drivers[i][j])
						{
							if (deviceIndices[i] == int(j))
							{
								sm.setDriver(i, drivers[i][j]);
							}
						}
					}
				}
				sm.enterMainMenuState();
			}
			else if (player != oldPlayer)
			{
				updatePlayer();
			}
			else if (device != oldDevice)
			{
				updateDevice();
			}
			else
			{
				for (int i = 0; i < 4; ++i)
				{
					if (enterControlMode[i])
					{
						menu.setActiveColor(STD_ACTIVE_COLOR2);
						enterControlMode[i] = false;
						mode = i;
						break;
					}
				}
			}
		}
		else if (key == SDLK_ESCAPE)
		{
			StateManager::getInstance().enterMainMenuState();
		}
	}
	else if (key == SDLK_ESCAPE)
	{
		menu.setActiveColor(STD_ACTIVE_COLOR);
		mode = -1;
	}
}

bool ControllerSettings::onEvent(const SDL_Event& event)
{
	if (mode != -1)
	{
		int i = player - 1;
		Driver* driver = drivers[i][deviceIndices[i]];
		if (driver->setControl(event, mode))
		{
			menu.setActiveColor(STD_ACTIVE_COLOR);
			updateControl(mode);
			mode = -1;
			return true;
		}
	}
	return false;
}

void ControllerSettings::updatePlayer()
{
	device = deviceIndices[player - 1];
	menu.notifyValueUpdate(&device);
	updateDevice();
}

void ControllerSettings::updateDevice()
{
	int i = player - 1;
	deviceIndices[i] = device;
	Driver*& driver = drivers[i][device];
	if (!driver)
	{
		driver = StateManager::getInstance().getDevice(device)->createDriver();
		driver->setDeviceIndex(device);
	}
	StateManager::getInstance().setDriver(i, driver);
	for (int i = 0; i < 4; ++i)
	{
		updateControl(i);
	}
}

void ControllerSettings::updateControl(int i)
{
	Driver* driver = drivers[player - 1][deviceIndices[player - 1]];
	menu.setLabel(2 + i, controlLabel[i] + driver->getControlName(i));
}
