/*
 * Copyright 2010 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Map.h"
#include <algorithm>
#include <cassert>
#include <iostream>

const int MIN_MAP_ROAD_COUNT = 10;

void Map::generate()
{
	// We generate a track until it is big enough. 
	// In majority of the cases we only need to try once.
	do
	{
		makeTrack();
	} while (getRoadCount() < MIN_MAP_ROAD_COUNT);
	
	// put track in the center of the map
	center();
}

void Map::zeroFill()
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			map[y][x] = 0;
		}
	}
}

int Map::getRoadCount() const
{
	int count = 0;
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			if (map[y][x] != 0)
			{
				count++;
			}
		}
	}
	return count;
} 
const int* Map::operator[](int y) const
{
	return map[y];
}

void Map::print() const
{	
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			switch (map[y][x])
			{
				case ROAD_HORIZONTAL: std::cout << "═"; break;
				case ROAD_VERTICAL: std::cout << "║"; break;
				case ROAD_LEFT_UP: std::cout << "╝"; break;
				case ROAD_LEFT_DOWN: std::cout << "╗"; break;
				case ROAD_RIGHT_UP: std::cout << "╚"; break;
				case ROAD_RIGHT_DOWN: std::cout << "╔"; break;
				case 0: std::cout << "·"; break;
				default: std::cout << "?"; break;
			}
		}
		std::cout << '\n';
	}
	std::cout << std::endl;
}

void Map::makeTrack()
{
	 // make sure the map is empty before we start
	zeroFill();
	
	bool verticalSplit = std::rand() & 1;
	
	if (verticalSplit)
	{
		int splitX = std::rand() % (MAX_MAP_X -1) + 1;
		
		int connectionY1 = std::rand() % MAX_MAP_Y;
		int connectionY2 = std::rand() % (MAX_MAP_Y - 3);
		if (connectionY2 >= connectionY1 - 1)
		{
			connectionY2 += 3;
		}
		
		makeConnection(0, splitX - 1, 0, MAX_MAP_Y - 1, splitX - 1, connectionY1, RIGHT, splitX - 1, connectionY2, RIGHT);
		makeConnection(splitX, MAX_MAP_X - 1, 0, MAX_MAP_Y - 1, splitX, connectionY1, LEFT, splitX, connectionY2, LEFT);
	}
	else
	{
		int splitY = std::rand() % (MAX_MAP_Y -1) + 1;
		
		int connectionX1 = std::rand() % MAX_MAP_X;
		int connectionX2 = std::rand() % (MAX_MAP_X - 3);
		if (connectionX2 >= connectionX1 - 1)
		{
			connectionX2 += 3;
		}
		makeConnection(0, MAX_MAP_X - 1, 0, splitY - 1, connectionX1, splitY - 1, DOWN, connectionX2, splitY - 1, DOWN);
		makeConnection(0, MAX_MAP_X - 1, splitY, MAX_MAP_Y - 1, connectionX1, splitY, UP, connectionX2, splitY, UP);
	}
}

void Map::makeConnection(int lowerX, int upperX, int lowerY, int upperY, int connectionX1, int connectionY1, int connectionDir1, int connectionX2, int connectionY2, int connectionDir2)
{
	if (connectionX1 == connectionX2 && connectionY1 == connectionY2)
	{
		// ORing two directions gives the connection
		map[connectionY1][connectionX1] = connectionDir1 | connectionDir2;
	}
	else
	{
		bool verticalSplit;
		if (connectionX1 == connectionX2)
		{
			verticalSplit = false;
		}
		else if (connectionY1 == connectionY2)
		{
			verticalSplit = true;
		}
		else 
		{
			verticalSplit = std::rand() & 1;
		}
		
		if (verticalSplit)
		{
			// make sure we know connection 1 is to the left
			// and connection 2 to the right
			if (connectionX1 > connectionX2)
			{
				std::swap(connectionX1, connectionX2);
				std::swap(connectionY1, connectionY2);
				std::swap(connectionDir1, connectionDir2);
			}

			// calculate the spliting plane
			int splitX = rand() % (connectionX2 - connectionX1) + connectionX1 + 1;

			int splitConnectionY = rand() % (upperY - lowerY + 1) + lowerY;
			
			// left side
			makeConnection(lowerX, splitX - 1,			// lo/up x
			  lowerY, upperY,							 // lo/up y
			  connectionX1, connectionY1, connectionDir1, // con. 1
			  splitX - 1, splitConnectionY, RIGHT);	   // con. 2

			// right side
			makeConnection(splitX, upperX,				// lo/up x
			  lowerY, upperY,							 // lo/up y
			  connectionX2, connectionY2, connectionDir2, // con. 1
			  splitX, splitConnectionY, LEFT);			// con. 2
		}
		else
		{
			// make sure we know connection 1 is over connection 2
			if (connectionY1 > connectionY2)
			{
				std::swap(connectionX1, connectionX2);
				std::swap(connectionY1, connectionY2);
				std::swap(connectionDir1, connectionDir2);
			}

			// calculate the spliting plane
			int splitY = rand() % (connectionY2 - connectionY1) + connectionY1 + 1;
			int splitConnectionX = rand() % (upperX - lowerX + 1) + lowerX;

			// up side
			makeConnection(lowerX, upperX,				// lo/up x
			  lowerY, splitY - 1,						 // lo/up y
			  connectionX1, connectionY1, connectionDir1, // con. 1
			  splitConnectionX, splitY - 1, DOWN);		// con. 2

			// down side
			makeConnection(lowerX, upperX,				// lo/up x
			  splitY, upperY,							 // lo/up y
			  connectionX2, connectionY2, connectionDir2, // con. 1
			  splitConnectionX, splitY, UP);			  // con. 2
		}
	}
}

void Map::center()
{
	int minX = MAX_MAP_Y;
	int minY = MAX_MAP_Y;
	int maxX = 0;
	int maxY = 0;
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			if (map[y][x] != 0)
			{
				minX = std::min(minX, x);
				minY = std::min(minY, y);
				maxX = std::max(maxX, x);
				maxY = std::max(maxY, y);
			}
		}
	}
	
	// The code below will work even if the map is empty
	
	int spaceLeft  = minX;
	int spaceRight = MAX_MAP_X - maxX - 1;
	int moveSpaceLeft = (spaceRight - spaceLeft) / 2;
	
	int spaceUp   = minY;
	int spaceDown = MAX_MAP_Y - maxY - 1;
	int moveSpaceUp = (spaceDown - spaceUp) / 2;
	
	if (moveSpaceLeft != 0 || moveSpaceUp != 0)
	{
		// move!
		Map temp = *this;
		zeroFill();
		for (int y = minY; y <= maxY; y++)
		{
			for (int x = minX; x <= maxX; x++)
			{
				map[(y + moveSpaceUp) % MAX_MAP_Y][(x + moveSpaceLeft) % MAX_MAP_X] = temp.map[y][x];
			}
		}
	}
}
