/*
 * Copyright 2011, 2013 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Menu.h"
#include "GameState.h"
#include <cstdarg>
#include <cassert>

const Color STD_INACTIVE_COLOR = {0x00, 0x00, 0x00};
const Color STD_ACTIVE_COLOR = {0x20, 0xFF, 0x20};
const Color STD_ACTIVE_COLOR2 = {0xFF, 0x20, 0x20};

Menu::Menu()
:	selectedItem(0),
	activeColor(STD_ACTIVE_COLOR),
	inactiveColor(STD_INACTIVE_COLOR)
{
}

void Menu::addItem(const std::string& label, int* value, int min, int max, const char* firstStrValue, ...)
{
	assert(value);
	assert(*value >= min);
	assert(*value <= max);
	
	Item item;
	item.label = label + ":";
	item.fireAction = 0;
	item.value = value;
	item.min = min;
	item.max = max;
	
	if (firstStrValue)
	{
		item.wrap = true;
		std::va_list vl;
		va_start(vl,firstStrValue);
		item.strValues.push_back(firstStrValue);
		for (int i = min + 1; i <= max; i++)
		{
			firstStrValue = va_arg(vl, const char*);
			item.strValues.push_back(firstStrValue);
		}
		va_end(vl);
	}
	else
	{
		item.wrap = false;
		for (int i = min; i <= max; i++)
		{
			item.strValues.push_back(int2string(i));
		}
	}
	items.push_back(item);
}
void Menu::addItem(const std::string& label, bool* fireAction)
{
	Item item;
	item.label = label;
	item.fireAction = fireAction;
	item.value = 0;
	item.min = 0;
	item.max = 0;
	item.wrap = false;
	items.push_back(item);
	
}

void Menu::addItem(const std::string& label, bool* fireAction, int* value, int min, int max, const char* firstStrValue, ...)
{
	assert(value);
	assert(*value >= min);
	assert(*value <= max);
	
	Item item;
	item.label = label + ": ";
	item.fireAction = fireAction;
	item.value = value;
	item.min = min;
	item.max = max;
	
	if (firstStrValue)
	{
		item.wrap = true;
		std::va_list vl;
		va_start(vl,firstStrValue);
		item.strValues.push_back(firstStrValue);
		for (int i = min + 1; i <= max; i++)
		{
			firstStrValue = va_arg(vl, const char*);
			item.strValues.push_back(firstStrValue);
		}
		va_end(vl);
	}
	else
	{
		item.wrap = false;
		for (int i = min; i <= max; i++)
		{
			item.strValues.push_back(int2string(i));
		}
	}
	items.push_back(item);
}

void Menu::addItem(const std::string& label, bool* fireAction, int* value, const std::vector<std::string>& strValues)
{
	assert(value);
	assert(*value >= 0);
	assert(unsigned(*value) < strValues.size());
	
	Item item;
	item.label = label + ": ";
	item.fireAction = fireAction;
	item.value = value;
	item.min = 0;
	item.max = strValues.size() - 1;
	item.strValues = strValues;
	item.wrap = true;
	items.push_back(item);
}

void Menu::addRenderObjects(GameState* gs, int x, int y, int dx, int dy)
{
	std::vector<Item>::iterator it;
	for (it = items.begin(); it != items.end(); ++it)
	{
		updateItem(it - items.begin());
		it->renderObject.moveTo(x, y);
		gs->addRenderObject(it->renderObject);
		x += dx;
		y += dy;
	}
}

bool Menu::onKeyDown(SDLKey key)
{
	switch (key)
	{
	case SDLK_DOWN:
		next();
		return true;
	case SDLK_UP:
		prev();
		return true;
	case SDLK_LEFT:
		dec();
		return true;
	case SDLK_RIGHT:
		inc();
		return true;
	case SDLK_RETURN:
	case SDLK_KP_ENTER:
		action();
		return true;
	default:
		return false;
	}
}

void Menu::next()
{
	std::size_t prevItem = selectedItem++;
	if (selectedItem == items.size())
	{
		selectedItem = 0;
	}
	updateItem(prevItem);
	updateItem(selectedItem);
}
void Menu::prev()
{
	std::size_t prevItem = selectedItem--;
	if (selectedItem >= items.size())
	{
		selectedItem = items.size() - 1;
	}
	updateItem(prevItem);
	updateItem(selectedItem);
}
void Menu::action()
{
	if (items[selectedItem].fireAction)
	{
		*items[selectedItem].fireAction = true;
	}
}
void Menu::inc()
{
	Item& item = items[selectedItem];
	if (item.value)
	{
		if (++(*item.value) > item.max)
		{
			*item.value = (item.wrap ? item.min : item.max);
		}
		updateItem(selectedItem);
	}
}
void Menu::dec()
{
	Item& item = items[selectedItem];
	if (item.value)
	{
		if (--(*item.value) < item.min)
		{
			*item.value = (item.wrap ? item.max : item.min);
		}
		updateItem(selectedItem);
	}
}

void Menu::notifyValueUpdate(int* value)
{
	assert(value);
	for (std::size_t i = 0; i < items.size(); i++)
	{
		if (value == items[i].value)
		{
			assert(*value >= items[i].min);
			assert(*value <= items[i].max);
			updateItem(i);
			break;
		}
	}
}

void Menu::selectLastItem()
{
	std::size_t prevItem = selectedItem;
	selectedItem = items.size() - 1;
	updateItem(prevItem);
	updateItem(selectedItem);
}

bool Menu::activeItemValue(int* value) const
{
	return items[selectedItem].value == value;
}

void Menu::setActiveColor(const Color& color)
{
	activeColor = color;
	updateItem(selectedItem);
}

void Menu::setLabel(int i, const std::string& label)
{
	items[i].label = label;
	updateItem(i);
}


void Menu::updateItem(std::size_t i)
{
	Item& item = items[i];
	std::string text = item.label;
	if (item.value)
	{
		text += item.strValues[*item.value - item.min];
	}
	Surface s(text, BIG_FONT, (i == selectedItem ? activeColor : inactiveColor));
	item.renderObject.setSurface(s);
}

void Menu::reset(GameState* gs)
{
	std::vector<Item>::iterator it;
	for (it = items.begin(); it != items.end(); ++it)
	{
		gs->removeRenderObject(it->renderObject);
	}
	items.clear();
	selectedItem = 0;
}
