/*
 * Copyright 2011 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "RaceSetup.h"
#include <algorithm>
#include <cassert>

RaceSetup::RaceSetup(const std::vector<Driver*>& drivers, const std::vector<int>& playerCars, int laps, int tracks, bool tournament, bool carCollisions)
:	players(drivers.size()),
	tournament(tournament),
	carCollisions(carCollisions),
	laps(laps),
	tracks(tracks),
	currentTrack(0)
{
	assert(drivers.size() == playerCars.size());

	for (std::size_t i = 0; i < players.size(); i++)
	{
		players[i].car = playerCars[i];
		players[i].driver = drivers[i];
		players[i].points = 0;
		players[i].time = 0;
	}
}

bool RaceSetup::isTournament() const
{
	return tournament;
}

bool RaceSetup::carCollisionsEnabled() const
{
	return carCollisions;
}

int RaceSetup::numberOfLaps() const
{
	return laps;
}

int RaceSetup::numberOfPlayers() const
{
	return players.size();
}

void RaceSetup::setPlayerPoints(int player, int points)
{
	players[player].points = points;
}

int RaceSetup::getPlayerPoints(int player) const
{
	return players[player].points;
}

void RaceSetup::setPlayerTime(int player, int time)
{
	players[player].time = time;
}

int RaceSetup::getPlayerTime(int player) const
{
	return players[player].time;
}

int RaceSetup::getPlayerCar(int player) const
{
	return players[player].car;
}

int RaceSetup::getPlayerPosition(int player) const
{
	return players[player].pos;
}

void RaceSetup::givePoints()
{
	for (std::size_t i = 0; i < players.size(); i++)
	{
		players[i].points += players.size() - players[i].pos - 1;
	}
}

std::vector<Driver*> RaceSetup::getDrivers() const
{
	std::vector<Driver*> drivers(players.size());
	for (std::size_t i = 0; i < players.size(); i++)
	{
		drivers[i] = players[i].driver;
	}
	return drivers;
}

void RaceSetup::nextTrack()
{
	currentTrack++;
}

int RaceSetup::getTracks() const
{
	return tracks;
}

int RaceSetup::getCurrentTrack() const
{
	return currentTrack;
}


bool RaceSetup::pointsCmp(const Player& p1, const Player& p2)
{
	return p1.points > p2.points;
}

void RaceSetup::sortPlayersByPoints()
{
	std::sort(players.begin(), players.end(), pointsCmp);
	int prevPoints = -1;
	int pos = 0;
	for (std::size_t i = 0; i < players.size(); i++)
	{
		if (players[i].points != prevPoints)
		{
			prevPoints = players[i].points;
			pos = i;
		}
		players[i].pos = pos;
	}
}

bool RaceSetup::timeCmp(const Player& p1, const Player& p2)
{
	return p1.time < p2.time;
}

void RaceSetup::sortPlayersByTime()
{
	std::sort(players.begin(), players.end(), timeCmp);
	int prevTime = -1;
	int pos = 0;
	for (std::size_t i = 0; i < players.size(); i++)
	{
		if (players[i].time != prevTime)
		{
			prevTime = players[i].time;
			pos = i;
		}
		players[i].pos = pos;
	}
}

void RaceSetup::reversePlayerOrder()
{
	std::reverse(players.begin(), players.end());
}
