/*
 * Copyright 2010, 2011, 2012 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Map.h"
#include "RaceState.h"
#include "SurfaceBank.h"
#include "StateManager.h"
#include "ResultState.h"
#include "TileSet.h"
#include <algorithm>
#include <cassert>
#include <climits>

const Color statColor = {0x00, 0x00, 0x00};
const Uint32 TIME_STATE_DURATION = 800;

bool RaceState::CarProgress::operator<(const RaceState::CarProgress& cp) const
{
	return progress < cp.progress;
}

RaceState::RaceState(const Map& map, const RaceSetup& rs)
:   raceSetup(rs),
	track(raceSetup.numberOfLaps(), raceSetup.numberOfPlayers(), raceSetup.getDrivers(), map, raceSetup.carCollisionsEnabled()),
	raceTime(0),
	endRace(false),
	returnFromMenu(false),
	timeState(TIME_STATE_READY)
{
	readyRO.setSurface(SurfaceBank::getReadySetGo(TIME_STATE_READY));
	setRO.setSurface(SurfaceBank::getReadySetGo(TIME_STATE_SET));
	goRO.setSurface(SurfaceBank::getReadySetGo(TIME_STATE_GO));
	
	int rsgOffsetX = 0;
	int rsgOffsetY = 0;
	switch (track.getFinishDir())
	{
	case UP:
		rsgOffsetY = 1;
		break;
	case DOWN:
		rsgOffsetY = 22;
		break;
	case LEFT:
		rsgOffsetX = -21;
		rsgOffsetY = 12;
		break;
	case RIGHT:
		rsgOffsetX = 22;
		rsgOffsetY = 12;
		break;
	}

	readyRO.moveTo(track.getFinishX() * TILE_SIZE + rsgOffsetX, track.getFinishY() * TILE_SIZE + rsgOffsetY);
	setRO.moveTo(track.getFinishX() * TILE_SIZE + rsgOffsetX, track.getFinishY() * TILE_SIZE + rsgOffsetY);
	goRO.moveTo(track.getFinishX() * TILE_SIZE + rsgOffsetX, track.getFinishY() * TILE_SIZE + rsgOffsetY);
	
	for (int i = 0; i < 10; i++)
	{
		smallNumbers[i] = Surface(int2string(i), SMALL_FONT, statColor);
		bigNumbers[i] = Surface(int2string(i), BIG_FONT, statColor);
	}
	
	TileSet::getRandomTileSet().decorate(map, track.getFinishX(), track.getFinishY(), background);
	
	int numCars = raceSetup.numberOfPlayers();
	// reserve to avoid invalidated pointers with addRenderObject
	cars.reserve(numCars);

	for (int i = 0; i < numCars; i++)
	{
		Car car;
		car.pos = 0;
		car.lap = 1;
		car.finished = false;
		const Vector& v = track.getCarPosition(i);
		car.carRO.update(v.x - CAR_RADIUS, v.y - CAR_RADIUS, track.getCarRotation(i));
		car.carRO.setSurface(SurfaceBank::getCar(raceSetup.getPlayerCar(i)));
		car.posRO.moveTo(17 + i * 31, 226);
		car.lapRO.moveTo(17 + i * 31, 233);
		car.timeRO.setSurface(Surface("lap: /"+int2string(raceSetup.numberOfLaps()), SMALL_FONT, statColor));
		car.timeRO.moveTo(3 + i * 31, 233);
		cars.push_back(car);
		addRenderObject(cars[i].carRO);
		addRenderObject(cars[i].posRO);
		addRenderObject(cars[i].lapRO);
		addRenderObject(cars[i].timeRO);
		
		SurfaceBank::getCarStatBackground(raceSetup.getPlayerCar(i)).draw(background, i * 31, 224);
		updateLapRO(i);
		Surface("pos: /"+int2string(numCars), SMALL_FONT, statColor).draw(background, 3 + i * 31, 226);
	}
	
	for (int i = 0; i < 6; i++)
	{
		timeRO[i].setSurface(bigNumbers[0]);
		timeNumbers[i] = 0;
		addRenderObject(timeRO[i]);
	}
	
	Surface timeColonSurface(":", BIG_FONT, statColor);
	timeRO[0].moveTo(310, 228);
	timeRO[1].moveTo(304, 228);
	timeColonSurface.draw(background, 302, 228);
	timeRO[2].moveTo(296, 228);
	timeRO[3].moveTo(290, 228);
	timeColonSurface.draw(background, 288, 228);
	timeRO[4].moveTo(282, 228);
	timeRO[5].moveTo(276, 228);
	
	addRenderObject(readyRO);
	
	// Start the timer after all other initilization
	gameTime = SDL_GetTicks();
}

const int TIME_STEP = 20;

void RaceState::update()
{
	int numCars = raceSetup.numberOfPlayers();
	
	if (returnFromMenu)
	{
		if (endRace) // user want's to end the race
		{
			Uint32 maxRaceTime = raceTime + 480000; // max 8 min
			Uint32 maxRealTime = SDL_GetTicks() + 10000; // max 10 sec
			while (!track.finished())
			{
				track.update(TIME_STEP * 0.001);
				raceTime += TIME_STEP;
				for (int i = 0; i < numCars; i++)
				{
					if (!cars[i].finished)
					{
						int lap = track.getCarLap(i);
						if (lap > raceSetup.numberOfLaps())
						{
							raceSetup.setPlayerTime(i, raceTime);
							cars[i].finished = true;
						}
					}
				}
				if (raceTime >= maxRaceTime || SDL_GetTicks() >= maxRealTime)
				{
					// we can't wait any longer for the cars to finish
					for (int i = 0; i < numCars; i++)
					{
						if (!cars[i].finished)
						{
							raceSetup.setPlayerTime(i, INT_MAX); // DNF
						}
					}
					break;
				}
			}
		}
		else
		{
			gameTime = SDL_GetTicks();
		}
	}
	
	if (track.finished() || endRace)
	{
		raceSetup.nextTrack();
		StateManager::getInstance().enterResultState(raceSetup);
		return;
	}
	
	do
	{
		if (timeState >= TIME_STATE_GO)
		{
			track.update(TIME_STEP * 0.001);
		}
		gameTime += TIME_STEP;
		raceTime += TIME_STEP;
	} while (gameTime <= SDL_GetTicks());
	
	for (int i = 0; i < numCars; i++)
	{
		const Vector& v = track.getCarPosition(i);
		cars[i].carRO.update(v.x - CAR_RADIUS, v.y - CAR_RADIUS, track.getCarRotation(i));
		
		int lap = std::max(1, track.getCarLap(i));
		if (lap != cars[i].lap)
		{
			cars[i].lap = lap;
			updateLapRO(i);
		}
		carProgress[i].carId = i;
		carProgress[i].progress = track.getCarProgress(i);
	}
	
	// update position stat
	std::sort(carProgress, carProgress + numCars);
	for (int i = 0; i < numCars; i++)
	{
		int carId = carProgress[i].carId;
		int pos = numCars - i;
		if (cars[carId].pos != pos)
		{
			cars[carId].pos = pos;
			updatePosRO(carId);
		}
	}
	
	updateTimeState();
}

void RaceState::updateTimeState()
{
	switch (timeState)
	{
	case TIME_STATE_READY:
		if (raceTime > TIME_STATE_DURATION)
		{
			timeState = TIME_STATE_SET;
			raceTime = 0;
			removeRenderObject(readyRO);
			addRenderObject(setRO);
		}
		break;
	case TIME_STATE_SET:
		if (raceTime > TIME_STATE_DURATION)
		{
			timeState = TIME_STATE_GO;
			raceTime = 0;
			removeRenderObject(setRO);
			addRenderObject(goRO);
		}
		break;
	case TIME_STATE_GO:
		if (raceTime > TIME_STATE_DURATION)
		{
			timeState = TIME_STATE_TIME;
			removeRenderObject(goRO);
		}
		// fall through
	case TIME_STATE_TIME:
		updateTimeRO();
		break;
	}
}

int RaceState::sleepTime()
{
	return gameTime - SDL_GetTicks();
}

void RaceState::updateLapRO(int i)
{
	assert(cars[i].lap >= 0);
	
	if (cars[i].lap <= raceSetup.numberOfLaps())
	{
		cars[i].lapRO.setSurface(smallNumbers[cars[i].lap]);
	}
	else
	{
		cars[i].timeRO.moveTo(2 + i * 31, 233);
		cars[i].timeRO.setSurface(Surface(time2string(raceTime), SMALL_FONT, statColor));
		removeRenderObject(cars[i].lapRO);
		raceSetup.setPlayerTime(i, raceTime);
		cars[i].finished = true;
	}
}

void RaceState::updatePosRO(int i)
{
	assert(cars[i].pos >= 0 && cars[i].pos <= 9);
	cars[i].posRO.setSurface(smallNumbers[cars[i].pos]);
}

void RaceState::updateTimeRO()
{
	Uint32 newTimeNumbers[6] =
	{
		(raceTime / 10) % 10,
		(raceTime / 100) % 10,
		(raceTime / 1000) % 10,
		(raceTime / 10000) % 6,
		(raceTime / 60000) % 10,
		(raceTime / 600000) % 10
	};
	
	for (int i = 0; i < 6; i++)
	{
		if (newTimeNumbers[i] != timeNumbers[i])
		{
			timeRO[i].setSurface(bigNumbers[newTimeNumbers[i]]);
			timeNumbers[i] = newTimeNumbers[i];
		}
	}
}

void RaceState::onKeyDown(SDLKey key)
{
	switch (key)
	{
	case SDLK_ESCAPE:
		returnFromMenu = true;
		StateManager::getInstance().enterInRaceMenuState(endRace);
		break;
	default:
		break;
	}
}
