/*
 * Copyright 2010, 2011 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "RenderObject.h"
#include "Vector.h"
#include <cmath>

RenderObject::RenderObject()
:	updated(true),
	x(0),
	y(0),
	rotationAngle(0.0)
{
	prevRect.x1 = prevRect.y1 = prevRect.x2 = prevRect.y2 = 0;
}


RenderObject::RenderObject(const Surface& surface, int x, int y, double rotate)
:	updated(true),
	x(x),
	y(y),
	rotationAngle(rotate),
	surface(surface)
{
	prevRect.x1 = prevRect.y1 = prevRect.x2 = prevRect.y2 = 0;
}

void RenderObject::update(int newX, int newY, double rotate)
{
	if (newX != x || newY != y || rotate != rotationAngle)
	{
		x = newX;
		y = newY;
		rotationAngle = rotate;
		updated = true;
	}
}
void RenderObject::moveTo(int newX, int newY)
{
	if (newX != x || newY != y)
	{
		x = newX;
		y = newY;
		updated = true;
	}
}
void RenderObject::setRotation(double rotate)
{
	if (rotate != rotationAngle)
	{
		rotationAngle = rotate;
		updated = true;
	}
}
void RenderObject::setSurface(const Surface& s)
{
	surface = s;
	updated = true;
}

void RenderObject::updateRects(Rect& newRect, Rect& oldRect)
{
	if (rotationAngle == 0.0)
	{
		newRect.x1 = x;
		newRect.y1 = y;
		newRect.x2 = x + surface.width();
		newRect.y2 = y + surface.height();
	}
	else
	{
		Vector middle(surface.width() * 0.5, surface.height() * 0.5);
		
		Vector p1(0, 0);
		Vector p2(surface.width(), 0);
		Vector p3(0, surface.height());
		Vector p4(surface.width(), surface.height());
		
		p1.rotate(middle, rotationAngle);
		p2.rotate(middle, rotationAngle);
		p3.rotate(middle, rotationAngle);
		p4.rotate(middle, rotationAngle);
		
		double minX = std::min(std::min(std::min(p1.x, p2.x), p3.x), p4.x);
		double minY = std::min(std::min(std::min(p1.y, p2.y), p3.y), p4.y);;
		double maxX = std::max(std::max(std::max(p1.x, p2.x), p3.x), p4.x);
		double maxY = std::max(std::max(std::max(p1.y, p2.y), p3.y), p4.y);
		
		newRect.x1 = x + std::floor(minX);
		newRect.y1 = y + std::floor(minY);
		newRect.x2 = x + std::ceil(maxX);
		newRect.y2 = y + std::ceil(maxY);
	}
	
	oldRect = prevRect;
	prevRect = newRect;
	updated = false;
}
void RenderObject::getOldRect(Rect& rect)
{
	rect = prevRect;
}

bool RenderObject::isUpdated() const
{
	return updated;
}

void RenderObject::flagUpdated()
{
	updated = true;
}

void RenderObject::draw(Surface& dst)
{
	surface.draw(dst, x, y, rotationAngle);
}
