/*
 * Copyright 2011, 2013 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ResultState.h"
#include "SurfaceBank.h"
#include "StateManager.h"
#include <cmath>

static const Color textColor = {0xFF, 0xDD, 0x10};

ResultState::ResultState(const RaceSetup& rs)
:	raceSetup(rs),
	nextAction(false),
	quitAction(false),
	pointsAction(false),
	haveWinner(false),
	winningCarRotation(0),
	winnerTextTimer(0)
{
	if (raceSetup.isTournament())
	{
		menu.addItem("point table", &pointsAction);
		
		extraTitle = " (" + int2string(rs.getCurrentTrack()) + '/' + int2string(rs.getTracks()) + ')';
	}
	else
	{
		 menu.addItem("New Race", &nextAction);
	}
	menu.addItem("Main menu", &quitAction);
	background.fill(120, 124, 224);
	
	raceSetup.sortPlayersByTime();
	
	Surface("Race Results" + extraTitle, BIG_FONT, textColor).draw(background, 124, 20);
	
	for (int i = 0; i < raceSetup.numberOfPlayers(); ++i)
	{
		Surface carSurface = SurfaceBank::getCar(raceSetup.getPlayerCar(i));
		carSurface.draw(background, 121, 40 + 20 * i);
		Surface(time2string(raceSetup.getPlayerTime(i)), BIG_FONT, textColor).draw(background, 161, 40 + 20 * i);
	}
	
	menu.addRenderObjects(this, 135, 200, 0, 12);
}

void ResultState::update()
{
	if (haveWinner)
	{
		winningCarRO.setRotation(5 * std::sin(winningCarRotation));
		winningCarRotation += 0.03;
		
		if ((winnerTextTimer % 15) == 0)
		{
			addRenderObject(leftWinnerTextRO);
			addRenderObject(rightWinnerTextRO);
		}
		else if ((winnerTextTimer % 15) == 10)
		{
			removeRenderObject(leftWinnerTextRO);
			removeRenderObject(rightWinnerTextRO);
		}
		++winnerTextTimer;
	}
}
void ResultState::onKeyDown(SDLKey key)
{
	if (menu.onKeyDown(key))
	{
		if (nextAction)
		{
			// make so the cars start next track in reversed posion order
			raceSetup.sortPlayersByTime();
			raceSetup.reversePlayerOrder();
			
			StateManager::getInstance().enterRaceState(raceSetup);
		}
		else if (quitAction)
		{
			StateManager::getInstance().enterMainMenuState();
		}
		else if (pointsAction)
		{
			raceSetup.givePoints();
			raceSetup.sortPlayersByPoints();
			
			menu.reset(this);
			background.fill(120, 124, 224);
			int tracksLeft = raceSetup.getTracks() - raceSetup.getCurrentTrack();
			
			std::string title;
			if (tracksLeft > 0)
			{
				menu.addItem("Next Race", &nextAction);
				title = "Points Table" + extraTitle;
			}
			else
			{
				haveWinner = true;
				
				Surface winnerText("Winner!!!", SMALL_FONT, textColor);
				leftWinnerTextRO.setSurface(winnerText);
				rightWinnerTextRO.setSurface(winnerText);
				winningCarRO.setSurface(SurfaceBank::getCar(raceSetup.getPlayerCar(0)));
				
				leftWinnerTextRO.moveTo(83, 41);
				rightWinnerTextRO.moveTo(200, 41);
				winningCarRO.moveTo(135, 40);
				
				addRenderObject(winningCarRO);
				title = "Final Points Table";
			}
			menu.addItem("Main menu", &quitAction);
			
			Surface(title, BIG_FONT, textColor).draw(background, 125, 20);
			
			for (int i = 0; i < raceSetup.numberOfPlayers(); ++i)
			{
				if (i != 0 || !haveWinner)
				{
					Surface carSurface = SurfaceBank::getCar(raceSetup.getPlayerCar(i));
					carSurface.draw(background, 135, 40 + 20 * i);
				}
				Surface(int2string(raceSetup.getPlayerPoints(i)), BIG_FONT, textColor).draw(background, 175, 40 + 20 * i);
			}
			
			menu.addRenderObjects(this, 135, 200, 0, 12);
			StateManager::getInstance().updateWholeScreen();
			pointsAction = false;
		}
	}
}
