/*
 * Copyright 2010, 2011, 2012, 2013 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "StateManager.h"
#include "GameState.h"
#include "RenderObject.h"
#include "MainMenu.h"
#include "RacePreparationMenu.h"
#include "CarSelectionState.h"
#include "RaceState.h"
#include "InRaceMenu.h"
#include "ResultState.h"
#include "ControllerSettings.h"
#include "Map.h"
#include "Path.h"
#include "Driver.h"
#include "KeyboardDriver.h"
#include "KeyboardDevice.h"
#include "MouseDevice.h"
#include "JoystickDevice.h"
#include <fstream>
#include <cassert>


StateManager::StateManager()
:	currentState(0),
	stateChanged(false),
	prevState(0),
	screenZoom(1),
	screenPaddingX(0),
	screenPaddingY(0),
	fullscreen(false)
{
	const SDL_VideoInfo* videoInfo = SDL_GetVideoInfo();
	
	// set the fullscreen dimensions
	screenWidth[1] = videoInfo->current_w;
	screenHeight[1] = videoInfo->current_h;

	// set the window dimensions
	int maxZoom = std::min(videoInfo->current_w / GAME_WIDTH, videoInfo->current_h / GAME_HEIGHT);
	screenZoom = std::max(maxZoom - 1, 1);
	screenWidth[0] = GAME_WIDTH * screenZoom;
	screenHeight[0] = GAME_HEIGHT * screenZoom;

	updateVideoMode();
	
	screenBuffer = Surface(GAME_WIDTH, GAME_HEIGHT);
	GameState::background = Surface(GAME_WIDTH, GAME_HEIGHT);
	
	// Read input devices
	devices.push_back(new KeyboardDevice());
	devices.push_back(new MouseDevice());
	int nJoysticks = SDL_NumJoysticks();
	for (int i = 0; i < nJoysticks; ++i)
	{
		devices.push_back(new JoystickDevice(i));
	}

	// load human drivers
	Path bindingsPath(configPath() + "/keybindings");
	if (bindingsPath.exist())
	{
		std::ifstream in(bindingsPath.c_str());
		std::string deviceName;
		while (drivers.size() < 8 && std::getline(in, deviceName))
		{
			for (std::size_t i = 0; i < devices.size(); ++i)
			{
				if (devices[i]->getName() == deviceName)
				{
					Driver* driver = devices[i]->createDriver();
					driver->load(in);
					driver->setDeviceIndex(i);
					drivers.push_back(driver);
					break;
				}
			}
		}
	}

	for (std::size_t i = drivers.size(); i < 8; ++i)
	{
		Driver* driver = 0;
		switch (i)
		{
		case 0:
			driver = new KeyboardDriver(SDLK_UP, SDLK_DOWN, SDLK_LEFT, SDLK_RIGHT);
			break;
		case 1:
			driver = new KeyboardDriver(SDLK_w, SDLK_s, SDLK_a, SDLK_d);
			break;
		case 2:
			driver = new KeyboardDriver(SDLK_i, SDLK_k, SDLK_j, SDLK_l);
			break;
		case 3:
			driver = new KeyboardDriver(SDLK_t, SDLK_g, SDLK_f, SDLK_h);
			break;
		case 4:
			driver = new KeyboardDriver(SDLK_KP8, SDLK_KP5, SDLK_KP4, SDLK_KP6);
			break;
		case 5:
			driver = new KeyboardDriver(SDLK_HOME, SDLK_END, SDLK_DELETE, SDLK_PAGEDOWN);
			break;
		case 6:
			driver = new KeyboardDriver(SDLK_F4, SDLK_F3, SDLK_F1, SDLK_F2);
			break;
		case 7:
			driver = new KeyboardDriver(SDLK_F10, SDLK_F9, SDLK_F7, SDLK_F8);
			break;
		}
		driver->setDeviceIndex(0);
		drivers.push_back(driver);
	}
}

void StateManager::resize(int width, int height)
{
	screenWidth[fullscreen] = std::max(GAME_WIDTH, width);
	screenHeight[fullscreen] = std::max(GAME_HEIGHT, height);
	updateVideoMode();
}

void StateManager::updateVideoMode()
{
	// update the zoom
	screenZoom = std::min(screenWidth[fullscreen] / GAME_WIDTH, screenHeight[fullscreen] / GAME_HEIGHT);
	
	// update padding
	screenPaddingX = (screenWidth[fullscreen] - GAME_WIDTH * screenZoom) / 2;
	screenPaddingY = (screenHeight[fullscreen] - GAME_HEIGHT * screenZoom) / 2;
	
	// update SDL window
	Uint32 flags = SDL_SWSURFACE|(fullscreen?SDL_FULLSCREEN|SDL_NOFRAME:SDL_RESIZABLE);
	SDL_SetVideoMode(screenWidth[fullscreen], screenHeight[fullscreen], 0, flags);
	
	//updateWholeScreen();
	// this will have to be called afterwards if needed
}

void StateManager::updateScreen()
{
	// draw background
	GameState::background.draw(screenBuffer,0,0);
	
	std::vector<RenderObject*>& renderObjects = currentState->renderObjects;
	std::vector<RenderObject*>::iterator it;

	Rect rect1;
	Rect rect2;
	for (it = renderObjects.begin(); it != renderObjects.end(); ++it)
	{
		RenderObject* ro = *it;
		ro->draw(screenBuffer);
		if (ro->isUpdated())
		{
			ro->updateRects(rect1, rect2);
			addUpdateRect(rect1);
			addUpdateRect(rect2);
		}
	}

	std::list<Rect>::iterator rit;
	for (rit = updateRects.begin(); rit != updateRects.end(); ++rit)
	{
		SDL_Rect rect;
		
		rect.x = (rit->x1);
		rect.y = (rit->y1);
		rect.w = (rit->x2 - rit->x1);
		rect.h = (rit->y2 - rit->y1);
		
		zoom_blit(screenZoom, screenPaddingX, screenPaddingY, screenBuffer, &rect);
		rect.x = (rit->x1) * screenZoom + screenPaddingX;
		rect.y = (rit->y1) * screenZoom + screenPaddingY;
		rect.w = (rit->x2 - rit->x1) * screenZoom + screenPaddingX;
		rect.h = (rit->y2 - rit->y1) * screenZoom + screenPaddingY;
		updateSDLRects.push_back(rect);
	}

	if (!updateSDLRects.empty())
	{

		SDL_UpdateRects(SDL_GetVideoSurface(), updateRects.size(), &updateSDLRects[0]);
		updateSDLRects.clear();
		updateRects.clear();
	}
}

void StateManager::updateWholeScreen()
{
	// Fill screen with black
	SDL_FillRect(SDL_GetVideoSurface(), 0, 0);

	// draw background
	GameState::background.draw(screenBuffer,0,0);
	
	// draw render objects
	std::vector<RenderObject*>& renderObjects = currentState->renderObjects;
	std::vector<RenderObject*>::iterator it;
	Rect rect1;
	Rect rect2;
	for (it = renderObjects.begin(); it != renderObjects.end(); ++it)
	{
		RenderObject* ro = *it;
		ro->draw(screenBuffer);
		if (ro->isUpdated())
		{
			// we need to call updateRects even if we don't use the 
			// rects to update them.
			ro->updateRects(rect1, rect2);
		}
	}
	
	zoom_blit(screenZoom, screenPaddingX, screenPaddingY, screenBuffer, 0);
	
	SDL_Flip(SDL_GetVideoSurface());
}

void StateManager::addUpdateRect(Rect& rect)
{
	if (rect.x1 >= rect.x2 || rect.y1 >= rect.y2)
	{
		return;
	}
	if (rect.x1 < 0)
	{
		rect.x1 = 0;
	}
	if (rect.y1 < 0)
	{
		rect.y1 = 0;
	}
	if (rect.x2 > GAME_WIDTH)
	{
		rect.x2 = GAME_WIDTH;
	}
	if (rect.y2 > GAME_HEIGHT)
	{
		rect.y2 = GAME_HEIGHT;
	}
	
	
	std::list<Rect>::iterator it;
	for (it = updateRects.begin(); it != updateRects.end(); ++it)
	{
		if (overlap(rect, *it))
		{
			rect.combine(*it);

			updateRects.erase(it);
			
			// start loop from start
			// because the new rect can be much larger
			it = updateRects.begin();
		}
	}
	updateRects.push_front(rect);
}

Driver* StateManager::getDriver(int index)
{
	assert(static_cast<unsigned>(index) < drivers.size());
	return drivers[index];
}

InputDevice* StateManager::getDevice(int index)
{
	assert(static_cast<unsigned>(index) < devices.size());
	return devices[index];
}

std::vector<std::string> StateManager::getDeviceNames() const
{
	std::vector<std::string> names;
	for (std::size_t i = 0; i < devices.size(); ++i)
	{
		names.push_back(devices[i]->getName());
	}
	return names;
}

void StateManager::setDriver(int i, Driver* driver)
{
	drivers[i] = driver;
}

void StateManager::cleanUp()
{
	// save (and delete) human drivers
	Path bindingsPath(configPath() + "/keybindings");
	std::ofstream out(bindingsPath.c_str());
	for (std::size_t i = 0; i < drivers.size(); ++i)
	{
		drivers[i]->save(out);
		delete drivers[i];
	}
	
	for (std::size_t i = 0; i < devices.size(); ++i)
	{
		delete devices[i];
	}
}

StateManager& StateManager::getInstance()
{
	static StateManager instance;
	return instance;
}

void StateManager::run()
{
	SDL_Event event;
	while (currentState)
	{
		stateChanged = false;

		// handle events
		while (SDL_PollEvent(&event))
		{
			handleEvent(event);
			if (stateChanged)
			{
				break;
			}
		}
		if (stateChanged)
		{
			// this one is needed because the other check inside the event
			// loop doesn't break out of the main loop
			continue;
		}
		
		// update state
		currentState->update();
		if (stateChanged)
		{
			continue;
		}
		// render
		updateScreen();
		
		// sleep
		int st = currentState->sleepTime();
		if (st > 0)
		{
			SDL_Delay(st);
		}
	}
}

void StateManager::handleEvent(const SDL_Event& event)
{
	if (currentState->onEvent(event))
	{
		return;
	}
	
	switch(event.type)
	{
	case SDL_KEYDOWN:
		if (SDL_GetModState() & KMOD_ALT)
		{
			switch (event.key.keysym.sym)
			{
			case SDLK_F4:
				enterQuitState();
				return;
			case SDLK_RETURN:
				toggleFullscreen();
				return;
			default:
				;
			}
		}
		switch (event.key.keysym.sym)
		{
		case SDLK_F11:
			toggleFullscreen();
			break;
		default:
			currentState->onKeyDown(event.key.keysym.sym);
			if (!stateChanged)
			{
				currentState->onUnicode(event.key.keysym.unicode);
			}
		}
		break;
	case SDL_KEYUP:
		currentState->onKeyUp(event.key.keysym.sym);
		break;
	case SDL_QUIT:
		enterQuitState();
		break;
	case SDL_VIDEORESIZE: 
		resize(event.resize.w, event.resize.h);
		updateWholeScreen();
		break;
	case SDL_VIDEOEXPOSE: 
		updateWholeScreen();
		break;
	default:
		;
	}
}

void StateManager::setNewState(GameState* state)
{
	// we normally don't save previous state
	// use savePrevState() to do that
	if (currentState)
	{
		delete prevState;
		prevState = 0;
		delete currentState;
	}
	currentState = state;
	stateChanged = true;
	if (state)
	{
		updateWholeScreen();
	}
}

// It is easy to add if we want a real quit state
// in the end.
void StateManager::enterQuitState()
{
	setNewState(0);
}

void StateManager::enterMainMenuState()
{
	setNewState(new MainMenu());
}

void StateManager::enterRacePreparationMenuState(bool tournament)
{
	setNewState(new RacePreparationMenu(tournament));
}

void StateManager::enterCarSelectionState(int humanPlayers, int computerPlayers, int laps, int tracks, bool tournament, bool carCollisions, int difficulty)
{
	setNewState(new CarSelectionState(humanPlayers, computerPlayers, laps, tracks, tournament, carCollisions, difficulty));
}

void StateManager::enterRaceState(const RaceSetup& raceSetup)
{
	Map map;
	map.generate();
	setNewState(new RaceState(map, raceSetup));
}



void StateManager::enterInRaceMenuState(bool& endRace)
{
	savePrevState();
	screenToBackgound();
	setNewState(new InRaceMenu(endRace));
}

void StateManager::enterResultState(const RaceSetup& raceSetup)
{
	setNewState(new ResultState(raceSetup));
}

void StateManager::enterControllerSettings()
{
	setNewState(new ControllerSettings());
}

// When entering a new state we sometimes want to use the old content of
// the screen as a background. This is useful for ingame menus. Call this
// method before creating the state object to make use of this.
void StateManager::screenToBackgound()
{
	screenBuffer.draw(GameState::background, 0, 0);
}

void StateManager::enterPrevState()
{
	GameState::background = prevStatebackground;
	GameState* newState = prevState;
	prevState = 0;
	setNewState(newState);
}

void StateManager::savePrevState()
{
	prevStatebackground = GameState::background;
	prevState = currentState;
	currentState = 0;
}

void StateManager::toggleFullscreen()
{
	fullscreen = !fullscreen;
	updateVideoMode();
	updateWholeScreen();
}
