/*
 * Copyright 2012 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "TileSet.h"
#include "Map.h"
#include "Path.h"
#include <cstdlib>

const int MIN_TILESET = 9;

TileSet::TileSet(const Surface& surface)
{
	wallColor = surface.getColor(0, 0);
	
	int nTiles = std::max(surface.height() / TILE_SIZE, MIN_TILESET);
	tiles.reserve(nTiles);
	
	for (int n = 0; n < nTiles; n++)
	{
		tiles.push_back(surface.subSurface(0, n * TILE_SIZE, TILE_SIZE, TILE_SIZE));
	}
}

void TileSet::decorate(const Map& map, int finishX, int finishY, Surface& background) const
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			const Surface& tile = pickTile(map, x, y, finishX, finishY);

			int px = x * TILE_SIZE;
			int py = y * TILE_SIZE;
			tile.draw(background, px, py);
			
			if (map[y][x] == 0)
			{
				if (y > 0 && map[y-1][x] != 0)
				{
					for (int i = 0; i < TILE_SIZE; i++)
					{
						background.setColor(px + i, py, wallColor); 
					}
				}
				if (x < (MAX_MAP_X - 1) && map[y][x + 1] != 0)
				{
					for (int i = 0; i < TILE_SIZE; i++)
					{
						background.setColor(px + TILE_SIZE - 1, py + i, wallColor); 
					}
				}
				
				if (x < (MAX_MAP_X - 1) && y > 0 && map[y - 1][x + 1] != 0)
				{
					background.setColor(px + TILE_SIZE - 1, py, wallColor); 
				}
			}
		}
	}
}

const Surface& TileSet::pickTile(const Map& map, int x, int y, int finishX, int finishY) const
{
	if (x == finishX && y == finishY)
	{
		if (map[finishY][finishX] == ROAD_HORIZONTAL)
		{
			return tiles[6]; 
		}
		else
		{
			return tiles[7];
		}
	}
	switch (map[y][x])
	{
		case ROAD_HORIZONTAL:
			return tiles[0];
		case ROAD_VERTICAL:
			return tiles[1];
		case ROAD_LEFT_UP:
			return tiles[2];
		case ROAD_LEFT_DOWN:
			return tiles[3];
		case ROAD_RIGHT_UP:
			return tiles[4];
		case ROAD_RIGHT_DOWN:
			return tiles[5];
		default:
			return tiles[MIN_TILESET + std::rand() % (tiles.size() - MIN_TILESET + 1) - 1];
	}
}

static bool isPNG(const std::string& filename)
{
	return filename.length() > 4 && filename.compare(filename.length() - 4, 4, ".png") == 0;
}

const TileSet& TileSet::getRandomTileSet()
{
	static std::vector<TileSet> tileSets;
	if (tileSets.empty()) // initialize
	{
		std::vector<std::string> files = (dataPath() + "tilesets").files();
		tileSets.reserve(files.size());
		for (std::size_t i = 0; i < files.size(); ++i)
		{
			if (isPNG(files[i]))
			{
				tileSets.push_back(TileSet(Surface("tilesets/" + files[i])));
			}
		}
		
		// make sure tileSets is not empty
		if (tileSets.empty())
		{
			tileSets.push_back(TileSet(Surface()));
		}
	}

	return tileSets[std::rand() % tileSets.size()];
}


