/*
 * Copyright 2010 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Vector.h"
#include <cmath>

const Vector VECTOR_X(1.0, 0.0);
const Vector VECTOR_Y(0.0, 1.0);

Vector::Vector()
:	x(0.0),
	y(0.0)
{
}
Vector::Vector(double x, double y)
:	x(x),
	y(y)
{
}

Vector::Vector(const Vector& p)
:	x(p.x),
	y(p.y)
{
}

double Vector::length() const
{
	return std::sqrt(x*x + y*y);
}

double Vector::lengthSquared() const
{
	return x*x + y*y;
}

void Vector::rotate(double angle)
{
	double cosAngle = std::cos(angle);
	double sinAngle = std::sin(angle);
	double tx = x;
	double ty = y;
	x = tx * cosAngle - ty * sinAngle;
	y = tx * sinAngle + ty * cosAngle;
}

void Vector::rotate(const Vector& p, double angle)
{
	double cosAngle = std::cos(angle);
	double sinAngle = std::sin(angle);
	double tx = x - p.x;
	double ty = y - p.y;
	x = tx * cosAngle - ty * sinAngle + p.x;
	y = tx * sinAngle + ty * cosAngle + p.y;
}

void Vector::normalize()
{
	double len = length();
	if (len != 0.0)
	{
		x = x / len;
		y = y / len;
	}
}

double Vector::angle(const Vector& v1, const Vector& v2)
{
	double lenSqr1 = v1.lengthSquared();
	double lenSqr2 = v2.lengthSquared();
	
	// the zero vector is orthogonal to all other vectors
	if (lenSqr1 == 0.0 || lenSqr2 == 0.0)
	{
		return PI_HALF;
	}
	return std::atan2(v2.y, v2.x) - std::atan2(v1.y, v1.x);
}

double Vector::dot(const Vector& v1, const Vector& v2)
{
	return (v1.x*v2.x + v1.y*v2.y);
}

bool operator ==(Vector& p1, Vector& p2)
{
	return (p1.x == p2.x && p1.y == p2.y);
}

bool operator !=(Vector& p1, Vector& p2)
{
	return (p1.x != p2.x || p1.y != p2.y);
}

Vector operator +(const Vector& p1, const Vector& p2)
{
	return Vector(p1.x + p2.x, p1.y + p2.y);
}

Vector operator -(const Vector& p1, const Vector& p2)
{
	return Vector(p1.x - p2.x, p1.y - p2.y);
}
Vector operator *(const Vector& p, double k)
{
	return Vector(k * p.x, k * p.y);
}
Vector operator *(double k, const Vector& p)
{
	return Vector(k * p.x, k * p.y);
}
Vector operator /(const Vector& p, double k)
{
	return Vector(p.x / k, p.y / k);
}
Vector operator /(double k, const Vector& p)
{
	return Vector(k / p.x, k / p.y);
}
Vector operator -(const Vector& p)
{
	return Vector(-p.x, -p.y);
}
Vector operator +(const Vector& p)
{
	return p;
}

Vector& operator +=(Vector& p1, const Vector& p2)
{
	p1.x += p2.x;
	p1.y += p2.y;
	return p1;
}

Vector& operator -=(Vector& p1, const Vector& p2)
{
	p1.x -= p2.x;
	p1.y -= p2.y;
	return p1;
}
Vector& operator *=(Vector& p1, double k)
{
	p1.x *= k;
	p1.y *= k;
	return p1;
}
Vector& operator /=(Vector& p1, double k)
{
	p1.x /= k;
	p1.y /= k;
	return p1;
}
