/*
 * Copyright 2010 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstddef>
#include <cstring>
#include <cassert>
#include "SDL.h"
#include "Surface.h"

template <typename T>
static inline void zoomBlit(const SDL_Rect& srcArea, int zoom, Sint16 x, Sint16 y, SDL_Surface* source, SDL_Surface* destination)
{
	T *dst = (T*) destination->pixels;
	T *src = (T*) source->pixels;
	
	dst += x + srcArea.x * zoom + (srcArea.y * zoom + y) * destination->pitch / sizeof(T);
	src += srcArea.x + srcArea.y * source->pitch / sizeof(T);
	
	std::ptrdiff_t src_skip = source->pitch / sizeof(T) - srcArea.w;
	std::ptrdiff_t dst_skip = destination->pitch / sizeof(T) - srcArea.w * zoom;
	
	std::ptrdiff_t dst_rowskip = destination->pitch / sizeof(T);
	
	// blit!
	int h = srcArea.h;
	while (h--) {
		T *dstTemp = dst;
		T *end_of_line = src + srcArea.w;
		while (src != end_of_line) {
			for (int zx = 0; zx < zoom; zx++) {
				*dstTemp = *src;
				dstTemp++;
			}
			src++;
		}
		dstTemp += dst_skip;
		// copy above row
		for (int zy = 1; zy < zoom; zy++) {
			std::memcpy(dstTemp, dst, srcArea.w * zoom * sizeof (T));
			dstTemp += dst_rowskip;
		}
		src += src_skip;
		dst += dst_rowskip * zoom;
	}
}

static inline void zoomBlit24(const SDL_Rect& srcArea, int zoom, Sint16 x, Sint16 y, SDL_Surface* source, SDL_Surface* destination)
{
	Uint8 *dst = (Uint8*) destination->pixels;
	Uint8 *src = (Uint8*) source->pixels;
	
	dst += 3 * (x + srcArea.x * zoom) + (srcArea.y * zoom + y) * destination->pitch;
	src += 3 * srcArea.x + srcArea.y * source->pitch;
	
	std::ptrdiff_t src_skip = source->pitch - 3 * srcArea.w;
	std::ptrdiff_t dst_skip = destination->pitch - 3 * srcArea.w * zoom;
	
	std::ptrdiff_t dst_rowskip = destination->pitch;
	
	// blit!
	int h = srcArea.h;
	while (h--) {
		Uint8 *dstTemp = dst;
		Uint8 *end_of_line = src + 3 * srcArea.w;
		while (src != end_of_line) {
			for (int zx = 0; zx < zoom; zx++) {
				dstTemp[0] = src[0];
				dstTemp[1] = src[1];
				dstTemp[2] = src[2];
				dstTemp += 3;
			}
			src += 3;
		}
		dstTemp += dst_skip;
		// copy above row
		for (int zy = 1; zy < zoom; zy++) {
			std::memcpy(dstTemp, dst, srcArea.w * zoom * 3);
			dstTemp += dst_rowskip;
		}
		src += src_skip;
		dst += dst_rowskip * zoom;
	}
}
void zoom_blit(int zoom, int offsetX, int offsetY, const Surface& srcSurface, SDL_Rect* area)
{
	SDL_Surface* dest = SDL_GetVideoSurface();
	SDL_Surface* src = srcSurface.surface;
	
	assert(src != 0);
	assert(dest != 0);
	assert(src->format->BytesPerPixel == dest->format->BytesPerPixel);
	assert(offsetX >= 0);
	assert(offsetY >= 0);
	assert(zoom > 0);
	
	SDL_Rect srcArea;
	
	if (!area)
	{
		srcArea.w = src->w;
		srcArea.h = src->h;
		srcArea.x = 0;
		srcArea.y = 0;
	} 
	else
	{
		srcArea.w = area->w;
		srcArea.h = area->h;
		srcArea.x = area->x;
		srcArea.y = area->y;
	}
	int bpp = dest->format->BytesPerPixel;
	switch (bpp) {
	case 1:
		zoomBlit<Uint8>(srcArea,zoom,offsetX,offsetY,src,dest);
		break;
	case 2:
		zoomBlit<Uint16>(srcArea,zoom,offsetX,offsetY,src,dest);
		break;
	case 3:
		zoomBlit24(srcArea,zoom,offsetX,offsetY,src,dest);
		break;
	case 4:
		zoomBlit<Uint32>(srcArea,zoom,offsetX,offsetY,src,dest);
		break;
	}
}
